package chapter5;

import java.io.BufferedWriter;
import java.io.IOException;
import Jama.Matrix;

public class MultThresholdLib {
	
	// prints a matrix to a file
	public static void printMatrix(double[][] matrix, int rows, int cols, BufferedWriter out) throws IOException{
		for (int i = 0; i < rows; i++){
			for (int j = 0; j < cols; j++){
				out.write((int)matrix[i][j] + " ");
			}
			out.newLine();
		}
	}
	
	// prints a matrix to the console
	public static void printMatrixConsole(double[][] matrix, int rows, int cols){
		for (int i = 0; i < rows; i++){
			for (int j = 0; j < cols; j++){
				System.out.print((int)matrix[i][j]);
			}
			System.out.println();
		}
		System.out.println();
	}
	
	// computes n choose r
	public static int nCr(int n, int r){
		
		double ncr = 1;
		int num = 1;
		int den = 1;
		
		for (int i = 0; i < r; i++){
			num = num*(n-i);
			den = den*(r-i);
		}
		
		ncr = (double)num/(double)den;
		return Math.round(Math.round(ncr));
	}
	
	// computes the number of zero elements in a vector
	public static int no_zeroelts(double[] vector, int elts){
		
		int ctr = 0;
		
		for (int i = 0; i < elts; i++){
			if ((int)vector[i] == 0){
				ctr++;
			}
		}
		return ctr;
	}
	
	// computes the modulus of an integer
	public static int mod(int x, int n){
		
		x = x%n;
		if (x < 0){
			x = x + n;
		}
		return x;
	}
	
	// computes the modulus of a matrix
	public static Matrix matrixMod(Matrix M, int rows, int cols, int n){
		
		double[][] Marray, MmodArray;
		Matrix Mmod;
		Marray = M.getArray();
		
		MmodArray = new double[rows][cols];		
		for (int i = 0; i < rows; i++){
			for (int j = 0; j < cols; j++){
				MmodArray[i][j] = mod((int)Marray[i][j],n);
			}
		}
		Mmod = new Matrix(MmodArray);
		return Mmod;
	}
	
	// returns true if two matrices M and N are equal
	public static Boolean matrixEqual(double[][] M, double[][] N, int rows, int cols){
		
		int diff;
		
		for (int i = 0; i < rows; i++){
			for (int j = 0; j < cols; j++){
				diff = (int)(M [i][j]-N[i][j]);
				if (diff != 0){
					return false;
				}
			}
		}
		return true;
	}
	
	// computes the star product of two vectors
	public static double[] star(double[] a, double[] b, int elts, int n){
		
		double[] c = new double[elts];
		for (int i = 0; i < elts; i++){
			c[i] = mod((int)(a[i]*b[i]),n);
		}
		return c;     
	}
	
	// computes the star product of a vector and a matrix
	public static double[][] starVM(double[] c, double[][] M, int rows, int cols, int n){
		Matrix Mmatrix, NTmatrix;
		double[][] MT, NT, N;
		
		Mmatrix = new Matrix(M);
		MT = Mmatrix.transpose().getArray();
		NT = new double[cols][rows];
		for (int i = 0; i < cols; i++){
			NT[i] = star(c,MT[i],rows,n);
		}
		NTmatrix = new Matrix(NT);
		N = NTmatrix.transpose().getArray();
		return N;
	}
	
	// computes M* = (m*m||m*M'||M'*M')
	public static double[][] Mstar(double[][] M, int rows, int cols, int n){
		
		Matrix mstarTMatrix, mstarMatrix, MMatrix;
		double[][] mstar, mstarT, MT;
		
		int cols2 = Math.round(Math.round((double)cols/2));
		int colsMstar = (cols+1)*cols2;
		
		mstar = new double[rows][colsMstar];
		mstarMatrix = new Matrix(mstar);
		mstarT = mstarMatrix.transpose().getArray();
		
		MMatrix = new Matrix(M);
		MT = MMatrix.transpose().getArray();
		
		int index = 0;
		
		for (int i = 0; i < cols; i++){
			for (int j = i; j < cols; j++){
				mstarT[index] = star(MT[i],MT[j],rows,n);
				index++;
			}
		}
		
		mstarTMatrix = new Matrix(mstarT);
		mstar = mstarTMatrix.transpose().getArray();
		return mstar;
	}
		
	// returns true if r is a reconstruction vector for S_M
	public static Boolean isRec(double[] r, double[][] M, int rows, int cols, int n) throws IOException{
		
		int[] a = new int[cols];
		
		a[0] = 0;
		for (int j = 0; j < rows; j++){
			a[0] = a[0] + (int) (M[j][0]*r[j]);
		}
		if (mod(a[0],n) != 1){
			return false;
		}
		
		for (int i = 1; i < cols; i++){
			a[i] = 0;
			for (int j = 0; j < rows; j++){
				a[i] = a[i] + (int) (M[j][i]*r[j]);
			}
			if (mod(a[i],n) != 0){
				return false;
			}
		}
		
		return true;
	}
	
	// returns true if S_M is multiplicative
	public static Boolean isMult(double[][] M, double[][] r, int rows, int cols, int n) throws IOException{
		
		int no_r, cols2, colsMstar;
		double[][] mstar;
		
		mstar = Mstar(M,rows,cols,n);
		cols2 = Math.round(Math.round((double)cols/2));
		colsMstar = (cols+1)*cols2;
		
		no_r = (int)Math.pow((double)n, (double)rows);
		for (int i = 0; i < no_r; i++){
			if (isRec(r[i],mstar,rows,colsMstar,n)){
				return true;
			}
		}
		return false;
	}
	
	// returns true if S_M is homomorphic
	public static Boolean isHom(double[][] M, double[][] r, int rows, int cols, int n) throws IOException{
		
		int no_r;
		int colsMstar;
		double[][] mstar;
		
		mstar = Mstar(M,rows,cols,n);
		colsMstar = (cols+1)*cols/2;
		
		no_r = (int)Math.pow((double)n, (double)rows);
		for (int i = 0; i < no_r; i++){
			if (isRec(r[i],M,rows,cols,n) && isRec(r[i],mstar,rows,colsMstar,n)){
				return true;
			}
		}
		return false;
	}
	
	// returns true if C is hyper-invertible, C nx1
	public static Boolean isHypInv1(double[] C, int rows, int n){
		
		for (int i = 0; i < rows; i++){
			if (C[i] == 0){
				return false;
			}
		}
		return true;
	}

	//returns true if C is hyper-invertible, C nx2
	public static Boolean isHypInv2(double[][] C, int rows, int n){
		double[][] Cprime;
		Matrix CprimeMatrix;
		
		Cprime = new double[2][2];
		CprimeMatrix = new Matrix(Cprime);
		
		for (int i = 0; i < rows; i++){
			for (int j = 0; j < 2; j++){
				if (C[i][j] == 0){
					return false;
				}
			}
		}
		for (int i = 0; i < rows; i++){
			for (int j = i+1; j < rows; j++){
				Cprime[0][0] = C[i][0];
				Cprime[0][1] = C[i][1];
				Cprime[1][0] = C[j][0];
				Cprime[1][1] = C[j][1];
				if (mod((int)CprimeMatrix.det(),n)==0){
					return false;
				}
			}
		}
		return true;
	}
	
	// returns true if any subset of (t+1) rows of a nx(t+1) matrix is of full rank
	// adapted from http://docs.google.com/Doc?id=ddd8c4hm_5fkdr3b, 16.5.2009
	public static Boolean isThreshold(double[][] M, int rows, int cols, int n){
		double[][] Mprime;
		Matrix MprimeMatrix;
		int max, j, maxComb, slots, tokens, index;
		
		Mprime = new double[cols][cols];
		MprimeMatrix = new Matrix(Mprime);
				
		maxComb = nCr(rows,cols);
		for (int k = 0; k < maxComb; k++){
			slots = rows;
			tokens = cols;
			j = 0;
			index = k;
			for (int i = 0; i < slots; i++ ){
				if (tokens == 0) break;
				max = nCr(slots-(i+1), tokens-1);
				if (index < max){
					Mprime[j] = M[i];
					j++;
					tokens--;
				}
				else if (index >= max){
					index = index - max;
	            }
	        }
			if (mod((int)MprimeMatrix.det(),n)==0){
				return false;
			}
		}
		return true;
	}
}
